@@ -4,6 +4,7 @@ author: Julien Tachoires <julmon@gmail.com>
 license: PostgreSQL License
 
 Copyright (c) 2012 - 2019, Julien Tachoires
+Copyright (c) 2020, Dalibo
 
 Permission to use, copy, modify, and distribute this software and its
 documentation for any purpose, without fee, and without a written
@@ -86,6 +87,8 @@ class Data:
     write_bytes_delta = 0
     read_count_delta = 0
     write_count_delta = 0
+    refresh_dbsize = False
+    min_duration = 0
 
     def __init__(self,):
         """
@@ -100,6 +103,8 @@ class Data:
         self.write_bytes_delta = 0
         self.read_count_delta = 0
         self.write_count_delta = 0
+        self.refresh_dbsize = False
+        self.min_duration = 0
 
     def get_pg_version(self,):
         """
@@ -270,6 +275,12 @@ class Data:
         """
         Get current sum of transactions, total size and  timestamp.
         """
+        prev_total_size = "0"
+        if prev_db_infos is not None:
+            prev_total_size = prev_db_infos['total_size']
+
+        skip_dbsize = skip_sizes and (not self.refresh_dbsize)
+
         query = """
     SELECT
         EXTRACT(EPOCH FROM NOW()) AS timestamp,
@@ -280,8 +291,8 @@ class Data:
         pg_database
         {no_rds}
         """.format(
-            db_size="0" if skip_sizes else "SUM(pg_database_size(datname))",
-            no_rds="WHERE datname <> 'rdsadmin'" if using_rds else ''
+            db_size = prev_total_size if skip_dbsize else "SUM(pg_database_size(datname))",
+            no_rds = "WHERE datname <> 'rdsadmin'" if using_rds else ''
         )
         cur = self.pg_conn.cursor()
         cur.execute(query,)
@@ -306,12 +317,22 @@ class Data:
         """
         Get total of active connections.
         """
-        query = """
-        SELECT
-            COUNT(*) as active_connections
-        FROM pg_stat_activity
-        WHERE state = 'active'
-        """
+
+        if self.pg_num_version < 90200:
+            # prior to PostgreSQL 9.1, there was no state column
+            query = """
+    SELECT
+        COUNT(*) as active_connections
+    FROM pg_stat_activity
+    WHERE current_query NOT LIKE '<IDLE>%%'
+            """
+        else:
+            query = """
+    SELECT
+        COUNT(*) as active_connections
+    FROM pg_stat_activity
+    WHERE state = 'active'
+            """
 
         cur = self.pg_conn.cursor()
         cur.execute(query,)
@@ -319,12 +340,12 @@ class Data:
         active_connections = int(ret['active_connections'])
         return active_connections
 
-    def pg_get_activities(self,):
+    def pg_get_activities(self, duration_mode=1):
         """
         Get activity from pg_stat_activity view.
         """
-        if self.pg_num_version >= 100000:
-            # PostgreSQL 10 and more
+        if self.pg_num_version >= 110000:
+            # PostgreSQL 11 and more
             query = """
     SELECT
         pg_stat_activity.pid AS pid,
@@ -339,22 +360,26 @@ class Data:
             ELSE pg_stat_activity.client_addr::TEXT
             END
         AS client,
-        EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) AS duration,
+        EXTRACT(epoch FROM (NOW() - pg_stat_activity.{duration_column})) AS duration,
         CASE WHEN pg_stat_activity.wait_event_type IN ('LWLock', 'Lock', 'BufferPin') THEN true ELSE false END AS wait,
         pg_stat_activity.usename AS user,
         pg_stat_activity.state AS state,
         pg_stat_activity.query AS query,
-        pg_stat_activity.backend_type AS backend_type
+        pg_stat_activity.backend_type = 'parallel worker' AS is_parallel_worker
     FROM
         pg_stat_activity
     WHERE
         state <> 'idle'
         AND pid <> pg_backend_pid()
+        AND CASE WHEN %(min_duration)s = 0 THEN true
+            ELSE extract(epoch from now() - {duration_column}) > %(min_duration)s
+            END
     ORDER BY
-        EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) DESC
+        EXTRACT(epoch FROM (NOW() - pg_stat_activity.{duration_column})) DESC
             """
-        elif self.pg_num_version >= 90600:
-            # PostgreSQL prior to 10.0 and >= 9.6.0
+        elif self.pg_num_version >= 100000:
+            # PostgreSQL 10
+            # We assume a background_worker with a not null query is a parallel worker.
             query = """
     SELECT
         pg_stat_activity.pid AS pid,
@@ -369,19 +394,56 @@ class Data:
             ELSE pg_stat_activity.client_addr::TEXT
             END
         AS client,
-        EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) AS duration,
-        pg_stat_activity.wait_event IS NOT NULL AS wait,
+        EXTRACT(epoch FROM (NOW() - pg_stat_activity.{duration_column})) AS duration,
+        CASE WHEN pg_stat_activity.wait_event_type IN ('LWLock', 'Lock', 'BufferPin') THEN true ELSE false END AS wait,
         pg_stat_activity.usename AS user,
         pg_stat_activity.state AS state,
         pg_stat_activity.query AS query,
-        null AS backend_type
+        (pg_stat_activity.backend_type = 'background worker' AND pg_stat_activity.query IS NOT NULL) AS is_parallel_worker
     FROM
         pg_stat_activity
     WHERE
         state <> 'idle'
         AND pid <> pg_backend_pid()
+        AND CASE WHEN %(min_duration)s = 0 THEN true
+            ELSE extract(epoch from now() - {duration_column}) > %(min_duration)s
+            END
     ORDER BY
-        EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) DESC
+        EXTRACT(epoch FROM (NOW() - pg_stat_activity.{duration_column})) DESC
+            """
+        elif self.pg_num_version >= 90600:
+            # PostgreSQL prior to 10.0 and >= 9.6.0
+            # There is no way to see parallel workers
+            query = """
+    SELECT
+        pg_stat_activity.pid AS pid,
+        pg_stat_activity.application_name AS application_name,
+        CASE WHEN LENGTH(pg_stat_activity.datname) > 16
+            THEN SUBSTRING(pg_stat_activity.datname FROM 0 FOR 6)||'...'||SUBSTRING(pg_stat_activity.datname FROM '........$')
+            ELSE pg_stat_activity.datname
+            END
+        AS database,
+        CASE WHEN pg_stat_activity.client_addr IS NULL
+            THEN 'local'
+            ELSE pg_stat_activity.client_addr::TEXT
+            END
+        AS client,
+        EXTRACT(epoch FROM (NOW() - pg_stat_activity.{duration_column})) AS duration,
+        pg_stat_activity.wait_event IS NOT NULL AS wait,
+        pg_stat_activity.usename AS user,
+        pg_stat_activity.state AS state,
+        pg_stat_activity.query AS query,
+        false AS is_parallel_worker
+    FROM
+        pg_stat_activity
+    WHERE
+        state <> 'idle'
+        AND pid <> pg_backend_pid()
+        AND CASE WHEN %(min_duration)s = 0 THEN true
+            ELSE extract(epoch from now() - {duration_column}) > %(min_duration)s
+            END
+    ORDER BY
+        EXTRACT(epoch FROM (NOW() - pg_stat_activity.{duration_column})) DESC
             """
         elif self.pg_num_version >= 90200:
             # PostgreSQL prior to 9.6.0 and >= 9.2.0
@@ -399,19 +461,22 @@ class Data:
             ELSE pg_stat_activity.client_addr::TEXT
             END
         AS client,
-        EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) AS duration,
+        EXTRACT(epoch FROM (NOW() - pg_stat_activity.{duration_column})) AS duration,
         pg_stat_activity.waiting AS wait,
         pg_stat_activity.usename AS user,
         pg_stat_activity.state AS state,
         pg_stat_activity.query AS query,
-        null AS backend_type
+        false AS is_parallel_worker
     FROM
         pg_stat_activity
     WHERE
         state <> 'idle'
         AND pid <> pg_backend_pid()
+        AND CASE WHEN %(min_duration)s = 0 THEN true
+            ELSE extract(epoch from now() - {duration_column}) > %(min_duration)s
+            END
     ORDER BY
-        EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) DESC
+        EXTRACT(epoch FROM (NOW() - pg_stat_activity.{duration_column})) DESC
             """
         elif self.pg_num_version < 90200:
             # PostgreSQL prior to 9.2.0
@@ -430,25 +495,44 @@ class Data:
             ELSE pg_stat_activity.client_addr::TEXT
             END
         AS client,
-        EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) AS duration,
+        EXTRACT(epoch FROM (NOW() - pg_stat_activity.{duration_column})) AS duration,
         pg_stat_activity.waiting AS wait,
         pg_stat_activity.usename AS user,
-        pg_stat_activity.current_query AS query,
-        null AS backend_type
+	CASE
+            WHEN pg_stat_activity.current_query = '<IDLE> in transaction (aborted)' THEN 'idle in transaction (aborted)'
+            WHEN pg_stat_activity.current_query = '<IDLE> in transaction' THEN 'idle in transaction'
+            WHEN pg_stat_activity.current_query = '<IDLE>' THEN 'idle'
+            ELSE 'active'
+            END
+        AS state,
+        CASE
+           WHEN pg_stat_activity.current_query LIKE '<IDLE>%%' THEN 'None'
+           ELSE pg_stat_activity.current_query
+           END
+        AS query,
+        false AS is_parallel_worker
     FROM
         pg_stat_activity
     WHERE
         current_query <> '<IDLE>'
         AND procpid <> pg_backend_pid()
+        AND CASE WHEN %(min_duration)s = 0 THEN true
+            ELSE extract(epoch from now() - {duration_column}) > %(min_duration)s
+            END
     ORDER BY
-        EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) DESC
+        EXTRACT(epoch FROM (NOW() - pg_stat_activity.{duration_column})) DESC
             """
+
+        duration_column = self.get_duration_column(duration_mode)
+        query = query.format(duration_column=duration_column)
+
         cur = self.pg_conn.cursor()
-        cur.execute(query)
+        cur.execute(query, {'min_duration': self.min_duration})
         ret = cur.fetchall()
+
         return ret
 
-    def pg_get_waiting(self,):
+    def pg_get_waiting(self, duration_mode=1):
         """
         Get waiting queries.
         """
@@ -466,7 +550,7 @@ class Data:
         pg_locks.mode AS mode,
         pg_locks.locktype AS type,
         pg_locks.relation::regclass AS relation,
-        EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) AS duration,
+        EXTRACT(epoch FROM (NOW() - pg_stat_activity.{duration_column})) AS duration,
         pg_stat_activity.state as state,
         pg_stat_activity.query AS query
     FROM
@@ -475,8 +559,11 @@ class Data:
     WHERE
         NOT pg_catalog.pg_locks.granted
         AND pg_catalog.pg_stat_activity.pid <> pg_backend_pid()
+        AND CASE WHEN %(min_duration)s = 0 THEN true
+            ELSE extract(epoch from now() - {duration_column}) > %(min_duration)s
+            END
     ORDER BY
-        EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) DESC
+        EXTRACT(epoch FROM (NOW() - pg_stat_activity.{duration_column})) DESC
             """
         elif self.pg_num_version < 90200:
             query = """
@@ -493,24 +580,41 @@ class Data:
         pg_locks.mode AS mode,
         pg_locks.locktype AS type,
         pg_locks.relation::regclass AS relation,
-        EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) AS duration,
-        NULL AS state,
-        pg_stat_activity.current_query AS query
+        EXTRACT(epoch FROM (NOW() - pg_stat_activity.{duration_column})) AS duration,
+        CASE
+           WHEN pg_stat_activity.current_query = '<IDLE> in transaction (aborted)' THEN 'idle in transaction (aborted)'
+           WHEN pg_stat_activity.current_query = '<IDLE> in transaction' THEN 'idle in transaction'
+           WHEN pg_stat_activity.current_query = '<IDLE>' THEN 'idle'
+           ELSE 'active'
+           END
+        AS state,
+        CASE
+           WHEN pg_stat_activity.current_query LIKE '<IDLE>%%' THEN 'None'
+           ELSE pg_stat_activity.current_query
+           END
+        AS query
     FROM
         pg_catalog.pg_locks
         JOIN pg_catalog.pg_stat_activity ON(pg_catalog.pg_locks.pid = pg_catalog.pg_stat_activity.procpid)
     WHERE
         NOT pg_catalog.pg_locks.granted
         AND pg_catalog.pg_stat_activity.procpid <> pg_backend_pid()
+        AND CASE WHEN %(min_duration)s = 0 THEN true
+            ELSE extract(epoch from now() - {duration_column}) > %(min_duration)s
+            END
     ORDER BY
-        EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) DESC
+        EXTRACT(epoch FROM (NOW() - pg_stat_activity.{duration_column})) DESC
             """
+
+        duration_column = self.get_duration_column(duration_mode)
+        query = query.format(duration_column=duration_column)
+
         cur = self.pg_conn.cursor()
-        cur.execute(query)
+        cur.execute(query, {'min_duration': self.min_duration})
         ret = cur.fetchall()
         return ret
 
-    def pg_get_blocking(self,):
+    def pg_get_blocking(self, duration_mode=1):
         """
         Get blocking queries
         """
@@ -542,7 +646,7 @@ class Data:
             pg_stat_activity.datname,
             pg_stat_activity.usename,
             blocking.locktype,
-            EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) AS duration,
+            EXTRACT(epoch FROM (NOW() - pg_stat_activity.{duration_column})) AS duration,
             pg_stat_activity.state as state,
             blocking.relation::regclass AS relation
         FROM
@@ -557,6 +661,9 @@ class Data:
             JOIN pg_stat_activity ON (blocking.pid = pg_stat_activity.pid)
         WHERE
             blocking.granted
+            AND CASE WHEN %(min_duration)s = 0 THEN true
+                ELSE extract(epoch from now() - {duration_column}) > %(min_duration)s
+                END
         UNION ALL
         SELECT
             blocking.pid,
@@ -566,7 +673,7 @@ class Data:
             pg_stat_activity.datname,
             pg_stat_activity.usename,
             blocking.locktype,
-            EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) AS duration,
+            EXTRACT(epoch FROM (NOW() - pg_stat_activity.{duration_column})) AS duration,
             pg_stat_activity.state as state,
             blocking.relation::regclass AS relation
         FROM
@@ -584,6 +691,9 @@ class Data:
             JOIN pg_stat_activity ON (blocking.pid = pg_stat_activity.pid)
         WHERE
             blocking.granted
+            AND CASE WHEN %(min_duration)s = 0 THEN true
+                ELSE extract(epoch from now() - {duration_column}) > %(min_duration)s
+                END
         ) AS sq
     GROUP BY
         pid,
@@ -603,6 +713,7 @@ class Data:
             query = """
     SELECT
         pid,
+        appname,
         CASE
             WHEN LENGTH(datname) > 16
             THEN SUBSTRING(datname FROM 0 FOR 6)||'...'||SUBSTRING(datname FROM '........$')
@@ -614,8 +725,18 @@ class Data:
         mode,
         locktype AS type,
         duration,
-        state,
-        query
+        CASE
+           WHEN sq.query = '<IDLE> in transaction (aborted)' THEN 'idle in transaction (aborted)'
+           WHEN sq.query = '<IDLE> in transaction' THEN 'idle in transaction'
+           WHEN sq.query = '<IDLE>' THEN 'idle'
+           ELSE 'active'
+           END
+        AS state,
+        CASE
+           WHEN sq.query LIKE '<IDLE>%%' THEN 'None'
+           ELSE sq.query
+           END
+        AS query
     FROM
         (
         SELECT
@@ -625,7 +746,7 @@ class Data:
             blocking.mode,
             pg_stat_activity.datname,
             pg_stat_activity.usename,
-            blocking.locktype,EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) AS duration,
+            blocking.locktype,EXTRACT(epoch FROM (NOW() - pg_stat_activity.{duration_column})) AS duration,
             NULL AS state,
             blocking.relation::regclass AS relation
         FROM
@@ -640,6 +761,9 @@ class Data:
             JOIN pg_stat_activity ON (blocking.pid = pg_stat_activity.procpid)
         WHERE
             blocking.granted
+            AND CASE WHEN %(min_duration)s = 0 THEN true
+                ELSE extract(epoch from now() - {duration_column}) > %(min_duration)s
+                END
         UNION ALL
         SELECT
             blocking.pid,
@@ -649,7 +773,7 @@ class Data:
             pg_stat_activity.datname,
             pg_stat_activity.usename,
             blocking.locktype,
-            EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) AS duration,
+            EXTRACT(epoch FROM (NOW() - pg_stat_activity.{duration_column})) AS duration,
             NULL AS state,
             blocking.relation::regclass AS relation
         FROM
@@ -667,6 +791,9 @@ class Data:
             JOIN pg_stat_activity ON (blocking.pid = pg_stat_activity.procpid)
         WHERE
             blocking.granted
+            AND CASE WHEN %(min_duration)s = 0 THEN true
+                ELSE extract(epoch from now() - {duration_column}) > %(min_duration)s
+                END
         ) AS sq
     GROUP BY
         pid,
@@ -682,8 +809,12 @@ class Data:
     ORDER BY
         duration DESC
             """
+
+        duration_column = self.get_duration_column(duration_mode)
+        query = query.format(duration_column=duration_column)
+
         cur = self.pg_conn.cursor()
-        cur.execute(query)
+        cur.execute(query, {'min_duration': self.min_duration})
         ret = cur.fetchall()
         return ret
 
@@ -701,6 +832,16 @@ class Data:
             return True
         return False
 
+    def get_duration_column(self, duration_mode=1):
+        if duration_mode not in (1, 2, 3):
+            duration_mode = 1
+        return ['query_start', 'xact_start', 'backend_start'][duration_mode-1]
+
+    def get_duration_mode_name(self, duration_mode=1):
+        if duration_mode not in (1, 2, 3):
+            duration_mode = 1
+        return ['query', 'transaction', 'backend'][duration_mode-1]
+
     def get_duration(self, duration):
         """
         Returns 0 if the given duration is negative
@@ -760,7 +901,7 @@ class Data:
                 process.set_extra('io_wait',
                     self.__sys_get_iow_status(psproc.status_iow()))
                 process.set_extra('psutil_proc', psproc)
-                process.set_extra('backend_type', query['backend_type'])
+                process.set_extra('is_parallel_worker', query['is_parallel_worker'])
                 process.set_extra('appname', query['application_name'])
 
                 processes[process.pid] = process
@@ -827,3 +968,9 @@ class Data:
         Get load average
         """
         return os.getloadavg()
+
+    def set_refresh_dbsize(self, refresh_dbsize):
+        """
+        Set self.refresh_dbsize
+        """
+        self.refresh_dbsize = refresh_dbsize
