@@ -4,6 +4,7 @@ author: Julien Tachoires <julmon@gmail.com>
 license: PostgreSQL License
 
 Copyright (c) 2012 - 2019, Julien Tachoires
+Copyright (c) 2020, Dalibo
 
 Permission to use, copy, modify, and distribute this software and its
 documentation for any purpose, without fee, and without a written
@@ -67,6 +68,11 @@ PGTOP_TRUNCATE =        1
 PGTOP_WRAP_NOINDENT =   2
 PGTOP_WRAP =            3
 
+# Duration mode
+PGTOP_DURATION_QUERY =          1
+PGTOP_DURATION_TRANSACTION =    2
+PGTOP_DURATION_BACKEND =        3
+
 # Cancel/Terminate backend
 PGTOP_SIGNAL_CANCEL_BACKEND =    ord('c')
 PGTOP_SIGNAL_TERMINATE_BACKEND = ord('k')
@@ -390,6 +396,8 @@ class UI:
         self.fs_blocksize = 4096
         # Output file
         self.output = None
+        # Parsed Cmd line options
+        self.options = None
 
     def set_verbose_mode(self, verbose_mode):
         """
@@ -403,6 +411,18 @@ class UI:
         """
         return self.verbose_mode
 
+    def set_duration_mode(self, duration_mode):
+        """
+        Set self.duration_mode
+        """
+        self.duration_mode = duration_mode
+
+    def get_duration_mode(self,):
+        """
+        Get self.duration_mode
+        """
+        return self.duration_mode
+
     def set_is_local(self, is_local):
         """
         Set self.is_local
@@ -729,6 +749,11 @@ class UI:
     def set_output(self, output):
         self.output = output
 
+    def set_options(self, options):
+        self.options = options
+        if self.data:
+            self.data.min_duration = options.minduration
+
     def clean_str(self, string):
         """
         Strip and replace some special characters.
@@ -1217,6 +1242,12 @@ class UI:
             if self.verbose_mode > 3:
                 self.verbose_mode = 1
             do_refresh = True
+        # change duration mode
+        if key == ord('T'):
+            self.duration_mode += 1
+            if self.duration_mode > 3:
+                self.duration_mode = 1
+            do_refresh = True
         # turn off/on colors
         if key == ord('C'):
             if self.color is True:
@@ -1240,9 +1271,15 @@ class UI:
         if key == ord('t') and self.sort != 't':
             self.sort = 't'
             known = True
-        if key == ord('+') and self.refresh_time < 3:
+        if key == ord('+') and self.refresh_time < 5 and self.refresh_time >= 1:
             self.refresh_time += 1
             do_refresh = True
+        if key == ord('+') and self.refresh_time < 1:
+            self.refresh_time = 1
+            do_refresh = True
+        if key == ord('-') and self.refresh_time == 1:
+            self.refresh_time = 0.5
+            do_refresh = True
         if key == ord('-') and self.refresh_time > 1:
             self.refresh_time -= 1
             do_refresh = True
@@ -1250,6 +1287,11 @@ class UI:
         if key == ord('R'):
             known = True
 
+        # Refresh db size
+        if key == ord('D'):
+            self.data.set_refresh_dbsize(True)
+            do_refresh = True
+
         if key == ord('u'):
             self.__empty_pid_yank()
             known = True
@@ -1292,7 +1334,7 @@ class UI:
                         disp_proc)
 
         # poll postgresql activity
-        queries =  self.data.pg_get_activities()
+        queries =  self.data.pg_get_activities(self.duration_mode)
         self.pid = []
         if self.is_local:
             # get resource usage for each process
@@ -1369,7 +1411,7 @@ class UI:
                         'duration': self.data.get_duration(proc.duration),
                         'wait': proc.wait,
                         'io_wait': proc.get_extra('io_wait'),
-                        'backend_type': proc.get_extra('backend_type')
+                        'is_parallel_worker': proc.get_extra('is_parallel_worker')
                     })
 
                 except psutil.NoSuchProcess:
@@ -1481,7 +1523,7 @@ class UI:
         if (k == curses.KEY_F1 or k == ord('1')):
             self.mode = 'activities'
             curses.flushinp()
-            queries = self.data.pg_get_activities()
+            queries = self.data.pg_get_activities(self.duration_mode)
             procs = self.data.sys_get_proc(queries, self.is_local)
             return self.__poll_activities(0, flag, indent, procs)
         # Waiting queries
@@ -1502,6 +1544,12 @@ class UI:
             if self.verbose_mode > 3:
                 self.verbose_mode = 1
             do_refresh = True
+        # change duration mode
+        if k == ord('T'):
+            self.duration_mode += 1
+            if self.duration_mode > 3:
+                self.duration_mode = 1
+            do_refresh = True
         # turnoff/on colors
         if k == ord('C'):
             if self.color is True:
@@ -1559,17 +1607,17 @@ class UI:
 
         # poll postgresql activity
         if self.mode == 'waiting':
-            queries =  self.data.pg_get_waiting()
+            queries =  self.data.pg_get_waiting(self.duration_mode)
         else:
-            queries =  self.data.pg_get_blocking()
+            queries =  self.data.pg_get_blocking(self.duration_mode)
 
         new_procs = {}
         for query in queries:
             new_procs[query['pid']] = query
-            new_procs[query['pid']][6] = \
+            new_procs[query['pid']]['duration'] = \
                 self.data.get_duration(query['duration'])
 
-        # return processes sorted by query duration
+        # return processes sorted by query/transaction/backend duration
         disp_procs = sorted(
                         queries,
                         key=lambda q: q['duration'],
@@ -1710,6 +1758,11 @@ class UI:
                     self.lineno,
                     colno,
                     " - Ref.: %ss" % (self.refresh_time,))
+        if self.options.minduration > 0:
+            colno += self.__print_string(
+                        self.lineno,
+                        colno,
+                        " - Min. duration: %ss" % (self.options.minduration,))
         colno = 0
         self.lineno += 1
         colno += self.__print_string(
@@ -1742,6 +1795,15 @@ class UI:
                     colno,
                     "%11s" % (active_connections,),
                     self.__get_color(C_GREEN)|curses.A_BOLD)
+        colno += self.__print_string(
+                    self.lineno,
+                    colno,
+                    "        | Duration mode: ")
+        colno += self.__print_string(
+                    self.lineno,
+                    colno,
+                    "%11s" % (self.data.get_duration_mode_name(self.duration_mode),),
+                    self.__get_color(C_GREEN)|curses.A_BOLD)
 
         # If not local connection, don't get and display system informations
         if not self.is_local:
@@ -1851,7 +1913,7 @@ class UI:
                 self.lineno,
                 00,
                 "      +",
-                "increase refresh time (max:3)")
+                "increase refresh time (max:5s)")
         self.__display_help_key(
                 self.lineno,
                 45,
@@ -1862,7 +1924,7 @@ class UI:
                 self.lineno,
                 00,
                 "      -",
-                "decrease refresh time (min:1)")
+                "decrease refresh time (min:0.5s)")
         self.__display_help_key(
                 self.lineno,
                 45,
@@ -1874,6 +1936,17 @@ class UI:
                 00,
                 "      R",
                 "force refresh")
+        self.__display_help_key(
+                self.lineno,
+                45,
+                "      T",
+                "change duration mode")
+        self.lineno += 1
+        self.__display_help_key(
+                self.lineno,
+                00,
+                "      D",
+                "force refresh database size")
         self.lineno += 1
         self.__print_string(
                 self.lineno,
@@ -2087,7 +2160,17 @@ class UI:
             elif process['duration'] >= 60000:
                 ctime = "%s h" % str(int(process['duration'] / 3600))
 
-            if process['duration'] < 1:
+            if process['duration'] == None:
+                # When duration is not available
+                colno += self.__print_string(
+                            l_lineno,
+                            colno,
+                            "%9s" % "N/A",
+                            self.line_colors['time_green'][typecolor])
+            elif process['duration'] < 1:
+                if process['duration'] < 0:
+                    # Set duration to 0 if < 0
+                    process['duration'] = 0
                 colno += self.__print_string(
                             l_lineno,
                             colno,
@@ -2152,10 +2235,9 @@ class UI:
         dif = self.maxx - len(indent) - 1
 
         query = ''
-        if 'backend_type' in process and \
-                process['backend_type'] == 'background worker':
-
+        if process.get('is_parallel_worker'):
             query += '\_ '
+
         if self.verbose_mode == PGTOP_TRUNCATE:
             query += process['query'][:dif]
             colno += self.__print_string(
